﻿using System.Collections.Generic;
using System.Linq;

namespace TreeTraversal
{
    static public class TreeTraversal
    {
        // Delegates
        public delegate IEnumerable<T> EnumerateChildren<T>(T item);

        private struct DFSWalkState<T>
        {
            // this is an internal struct used when traversing the DOM
            // it preserves the state in the non-recursive, stack-based traversal of the DOM

            internal readonly T Node;
            internal bool Entered;
            internal int Depth;

            public DFSWalkState(T node, int depth)
            {
                this.Node = node;
                this.Entered = false;
                this.Depth = depth;
            }
        }

        /// <summary>
        // Walks a Node in a depth-first manner without recursion.
        // It returns a series of "events" that indicate one of three things:
        // - whether it has enters into a node
        // - whether it has exited from a node (i.e. it is finished with that container and its children)
        // - caller can control which children get entered via the enum_children method
        /// </summary>
        public static IEnumerable<DFSEvent<T>> DFSWalk<T>(T node, TreeTraversal.EnumerateChildren<T> enum_children)
        {
            var stack = new Stack<DFSWalkState<T>>();

            const int starting_depth = 0;
            var inital_state = new DFSWalkState<T>(node, starting_depth);
            // put the first item on the stack 
            stack.Push(inital_state);

            // reversbuf is used to optimize for adding children to the stack in reverse
            List<T> reversebuf = null;

            // As long as something is on the stack, we are not done
            while (stack.Count > 0)
            {
                var state = stack.Pop();

                if (state.Entered == false)
                {
                    var ev = DFSEvent<T>.CreateEnterEvent(state.Node, state.Depth);
                    yield return ev;

                    state.Entered = true;
                    stack.Push(state);

                    AddChildrenToStackInReverse(state, enum_children, stack, ref reversebuf);
                }
                else
                {
                    var ev = DFSEvent<T>.CreateExitEvent(state.Node, state.Depth);
                    yield return ev;
                }
            }
        }

        private static void AddChildrenToStackInReverse<T>(DFSWalkState<T> walkstate, TreeTraversal.EnumerateChildren<T> enum_children, Stack<DFSWalkState<T>> stack, ref List<T> reversebuf)
        {
            // Add the children in reverse
            // Optimize for the case when the children are just an IList<>

            var children = enum_children(walkstate.Node);

            if (children is IList<T>)
            {
                // If it is a list then just push the children in backwards
                var item_col = (IList<T>)children;
                for (int i = item_col.Count - 1; i >= 0; i--)
                {
                    var newstate = new DFSWalkState<T>(item_col[i], walkstate.Depth + 1);
                    stack.Push(newstate);
                }
            }
            else
            {
                // otherise store the items in the reversebuf temporarily for reversal
                if (reversebuf == null)
                {
                    reversebuf = new List<T>();
                }

                reversebuf.AddRange(children);

                for (int i = reversebuf.Count - 1; i >= 0; i--)
                {
                    var newstate = new DFSWalkState<T>(reversebuf[i], walkstate.Depth + 1);
                    stack.Push(newstate);
                }

                reversebuf.Clear();
            }
        }

        public static IEnumerable<T> PreOrder<T>(T root, EnumerateChildren<T> enum_children)
        {
            return DFSWalk(root, enum_children)
                .Where(ev => ev.HasEnteredNode)
                .Select(ev => ev.Node);
        }

        public static IEnumerable<T> PostOrder<T>(T root, EnumerateChildren<T> enum_children)
        {
            return DFSWalk(root, enum_children)
                .Where(ev => ev.HasExitedNode)
                .Select(ev => ev.Node);
        }

        public static IEnumerable<T> BreadFirst<T>(T root, EnumerateChildren<T> enum_children)
        {
            var q = new Queue<T>();
            q.Enqueue(root);

            while (q.Count > 0)
            {
                var n = q.Dequeue();
                yield return n;

                foreach (var c in enum_children(n))
                {
                    q.Enqueue(c);
                }
            }
        }
    }
}