﻿using System.Collections.Generic;
using System.Linq;

namespace Viziblr.Drawing
{
    public struct BezierSegment
    {
        private readonly Point _start;
        private readonly Point _handle1;
        private readonly Point _handle2;
        private readonly Point _end;

        public Point Start
        {
            get { return _start; }
        }

        public Point Handle1
        {
            get { return _handle1; }
        }

        public Point Handle2
        {
            get { return _handle2; }
        }

        public Point End
        {
            get { return _end; }
        }

        public Point[] ToPoints()
        {
            var points = new Point[4] { this._start, this._handle1, this._handle2, this._end };
            return points;
        }

        public BezierSegment(Point start, Point handle1, Point handle2, Point end)
        {
            this._start = start;
            this._handle1 = handle1;
            this._handle2 = handle2;
            this._end = end;
        }

        public BezierSegment(Point[] points)
        {
            if (points == null)
            {
                throw new System.ArgumentNullException("points");    
            }

            if (points.Length != 4)
            {
                throw new DrawingException("Must have exactly 4 points");
            }

            this._start = points[0];
            this._handle1 = points[1];
            this._handle2 = points[2];
            this._end = points[3];
        }

        public static IList<Point> Merge(IList<BezierSegment> segments, out int degree)
        {
            if (segments == null)
            {
                throw new System.ArgumentNullException("segments");
            }

            var points = new List<Point>(segments.Count * 4);
            int n = 0;
            foreach (var seg in segments)
            {
                if (n == 0)
                {
                    points.Add(seg._start);
                }

                points.Add(seg._handle1);
                points.Add(seg._handle2);
                points.Add(seg._end);
                n++;
            }

            degree = 3;
            return points;
        }

        public static BezierSegment[] FromArc(double startangle, double endangle)
        {
            double min_angle = 0;
            double max_angle = System.Math.PI * 2;

            double total_angle = endangle - startangle;

            if (total_angle == min_angle)
            {
                return null;
            }

            if (total_angle > max_angle)
            {
                endangle = startangle + max_angle;
            }

            var bez_arr = DrawingUtil.subdivide_arc_nicely(startangle, endangle)
                .Select<ArcSegment, BezierSegment>(a => get_bezier_points_for_small_arc(a.begin, a.end))
                .ToArray<BezierSegment>();

            return bez_arr;
        }

        internal static BezierSegment get_bezier_points_for_small_arc(double start_angle, double end_angle)
        {
            const double right_angle = System.Math.PI / 2;
            double total_angle = end_angle - start_angle;

            if (total_angle > right_angle)
            {
                throw new System.ArgumentOutOfRangeException("end_angle", "angle formed by start and end must <= right angle (pi/2)");
            }

            double theta = (end_angle - start_angle) / 2;
            double cos_theta = System.Math.Cos(theta);
            double sin_theta = System.Math.Sin(theta);

            var p0 = new Point(cos_theta, -sin_theta);
            var p1 = new Point((4 - cos_theta) / 3.0, ((1 - cos_theta) * (cos_theta - 3.0)) / (3 * sin_theta));
            var p2 = new Point(p1.X, -p1.Y);
            var p3 = new Point(p0.X, -p0.Y);

            var arc_bezier = new Point[] { p0, p1, p2, p3 }
                .Select(p => DrawingUtil.RotateAroundOrigin(p, theta + start_angle))
                .ToArray();

            return new BezierSegment(arc_bezier);
        }
    }
}