using System.Collections.Generic;
using System.Linq;

namespace Viziblr.Drawing
{
    public class NURBSCurve
    {
        public double[] Knots { get; set; }
        public Point[] ControlPoints { get; set; }
        public double[] Weights { get; set; }
        public int Degree { get; set; }

        public int Order
        {
            get { return this.Degree + 1; }
        }

        public NURBSCurve()
        {
            this.Degree = -1;
        }

        public NURBSCurve(Point[] points, double[] weights, double[] knots, int degree)
        {
            // TODO: verify number of points and degree matches

            if (points == null)
            {
                throw new System.ArgumentNullException("points");
            }

            if (points.Length < 1)
            {
                throw new System.ArgumentOutOfRangeException("points");
            }


            if (weights==null)
            {
                throw new System.ArgumentNullException("weights");
            }

            if (weights.Length < 1)
            {
                throw new System.ArgumentOutOfRangeException("weights");
            }


            if (weights.Length != points.Length)
            {
                throw new System.ArgumentException("Number of weights and points must be equal","weights");
            }

            this.ControlPoints = points;
            this.Weights = weights;
            this.Knots = knots;
            this.Degree = degree;
        }

        public static Viziblr.Drawing.NURBSCurve FromCircle(Point origin, double radius)
        {
            // from here: http://en.wikipedia.org/wiki/NURBS

            var curve = new NURBSCurve();
            curve.Knots = new double[]
                              {
                                  0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4
                              };
            var doubles = new double[]
                              {
                                  radius, 0,
                                  radius, radius,
                                  0, radius,
                                  -radius, radius,
                                  -radius, 0,
                                  -radius, -radius,
                                  0, -radius,
                                  radius, -radius,
                                  radius, 0
                              };

            curve.ControlPoints = Point.DoublesToPoints(doubles).Select( x => x + origin).ToArray();

            double sq = System.Math.Sqrt(2)/2.0;
            curve.Weights = new double[] {1, sq, 1, sq, 1, sq, 1, sq, 1};
            curve.Degree = 2;
            return curve;
        }

        public static NURBSCurve FromBezierCurve(BezierCurve bez_curve)
        {
            //TODO: verify expected number of points for degree

            if (bez_curve.Degree != 3)
            {
                string msg = "Degree must be 3 - support for other degrees not implemented";
                throw new Viziblr.Drawing.DrawingException(msg);
            }

            int kn = ((bez_curve.ControlPoints.Length - 1)/3) + 1;

            var knotlist = new List<double>(kn + 2);
            knotlist.Add(0);

            for (int i=0;i<kn;i++)
            {
                knotlist.Add(i);
                knotlist.Add(i);
                knotlist.Add(i);
            }

            knotlist.Add(kn - 1);

            var knots = knotlist.ToArray();

            var points = bez_curve.ControlPoints.ToArray();
            var weights = Enumerable.Repeat(1.0, points.Length).ToArray();
            var curve = new NURBSCurve(points, weights, knots, 3);

            return curve;
        }
    }
}