using System.Collections.Generic;

namespace Viziblr.Drawing
{
    public struct Rectangle
    {
        private readonly double _left;
        private readonly double _bottom;
        private readonly double _right;
        private readonly double _top;

        public Rectangle(double left, double bottom, double right, double top)
        {
            if (right < left)
            {
                throw new System.ArgumentException("left must be <=right");
            }

            if (top < bottom)
            {
                throw new System.ArgumentException("bottom must be <= top");
            }

            _left = left;
            _bottom = bottom;
            _right = right;
            _top = top;
        }

        public Rectangle(Point lowerleft, Point upperright)
        {
            if (upperright.X < lowerleft.X)
            {
                throw new System.ArgumentException("left must be <=right");
            }

            if (upperright.Y < lowerleft.Y)
            {
                throw new System.ArgumentException("bottom must be <= top");
            }

            _left = lowerleft.X;
            _bottom = lowerleft.Y;
            _right = upperright.X;
            _top = upperright.Y;
        }

        public Rectangle(Point lowerleft, Viziblr.Drawing.Size s)
        {
            if (s.Width < 0)
            {
                throw new System.ArgumentOutOfRangeException("s", "width must be non-negative");
            }

            if (s.Height < 0)
            {
                throw new System.ArgumentOutOfRangeException("s", "height must be non-negative");
            }

            _left = lowerleft.X;
            _bottom = lowerleft.Y;
            _right = lowerleft.X + s.Width;
            _top = lowerleft.Y + s.Height;
        }

        public static Rectangle FromCenterPoint(double x, double y, double w, double h)
        {
            if (w < 0)
            {
                throw new System.ArgumentOutOfRangeException("w", "width must be non-negative");
            }

            if (h < 0)
            {
                throw new System.ArgumentOutOfRangeException("h", "height must be non-negative");
            }

            var xradius = w/2.0;
            var yradius = h/2.0;
            var r = new Viziblr.Drawing.Rectangle(x - xradius, y - yradius, x + xradius, y + yradius);
            return r;
        }

        public static Rectangle FromCenterPoint(Viziblr.Drawing.Point p, double width, double height)
        {
            return FromCenterPoint(p.X, p.Y, width, height);
        }

        public void ToDoubles(out double left, out double bottom, out double right, out double top)
        {
            left = _left;
            right = _right;
            top = _top;
            bottom = _bottom;
        }

        public Point[] GetPointsForClosedPolygon()
        {
            Point[] points = {LowerLeft, LowerRight, UpperRight, UpperLeft, LowerLeft};
            return points;
        }

        public override string ToString()
        {
            string s = string.Format(System.Globalization.CultureInfo.InvariantCulture, "Rectangle({0},{1},{2},{3})",
                                     _left, _bottom, _right, _top);
            return s;
        }

        public Point LowerLeft
        {
            get { return new Point(_left, _bottom); }
        }

        public Point LowerRight
        {
            get { return new Point(_right, _bottom); }
        }

        public Point UpperLeft
        {
            get { return new Point(_left, _top); }
        }

        public Point UpperRight
        {
            get { return new Point(_right, _top); }
        }

        public Size Size
        {
            get { return new Size(Width, Height); }
        }

        public double Width
        {
            get { return _right - _left; }
        }

        public double Height
        {
            get { return _top - _bottom; }
        }

        public double DiagonalLength
        {
            get { return System.Math.Sqrt((Width*Width) + (Height*Height)); }
        }

        public Point CenterPoint
        {
            get { return new Point((_left + _right)/2.0, (_bottom + _top)/2.0); }
        }

        public double Left
        {
            get { return _left; }
        }

        public double Bottom
        {
            get { return _bottom; }
        }

        public double Right
        {
            get { return _right; }
        }

        public double Top
        {
            get { return _top; }
        }

        public static Rectangle operator +(Rectangle r, Point p)
        {
            return r.Add(p.X, p.Y);
        }

        public static Rectangle operator -(Rectangle r, Point p)
        {
            return r.Subtract(p.X, p.Y);
        }

        public static Rectangle operator *(Rectangle r, double s)
        {
            return r.Multiply(s, s);
        }

        public Rectangle Add(double dx, double dy)
        {
            var r2 = new Rectangle(_left + dx, _bottom + dy, _right + dx, _top + dy);
            return r2;
        }

        public Rectangle Subtract(double dx, double dy)
        {
            var r2 = new Rectangle(_left - dx, _bottom - dy, _right - dx, _top - dy);
            return r2;
        }

        public Rectangle Multiply(double sx, double sy)
        {
            var r2 = new Rectangle(_left*sx, _bottom*sy, _right*sx, _top*sy);
            return r2;
        }

        public static Rectangle FromBoundingBox(Rectangle r1, Rectangle r2)
        {
            double left = System.Math.Min(r1._left, r2._left);
            double bottom = System.Math.Min(r1._bottom, r2._bottom);
            double right = System.Math.Max(r1._right, r2._right);
            double top = System.Math.Max(r1._top, r2._top);

            var RECT = new Rectangle(left, bottom, right, top);
            return RECT;
        }

        public static Rectangle FromBoundingBox(Point p1, Point p2)
        {
            double left = System.Math.Min(p1.X, p2.X);
            double bottom = System.Math.Min(p1.Y, p2.Y);
            double right = System.Math.Max(p1.X, p2.X);
            double top = System.Math.Max(p1.X, p2.X);

            var RECT = new Rectangle(left, bottom, right, top);
            return RECT;
        }

        public static Rectangle FromBoundingBox(IEnumerable<Point> points)
        {
            Rectangle? result = DrawingUtil.TryGetBoundingBox(points);
            if (!result.HasValue)
            {
                string msg = "Failed to create bounding box from points";
                throw new System.ArgumentException(msg, "points");
            }
            return result.Value;
        }
    }
}