namespace Viziblr.Drawing
{
    public struct Size
    {
        private readonly double _width;
        private readonly double _height;

        public Size(double width, double height)
        {
            if (width<0.0)
            {
                throw new System.ArgumentOutOfRangeException("width");
            }
            if (height< 0.0)
            {
                throw new System.ArgumentOutOfRangeException("height");
            }
            _width = width;
            _height = height;
        }

        public double Width
        {
            get { return _width; }
        }

        public double Height
        {
            get { return _height; }
        }

        public override string ToString()
        {
            string s = string.Format(System.Globalization.CultureInfo.InvariantCulture,"Size({0}x{1})", _width, _height);
            return s;
        }

        public bool GreaterThanOrEqualTo(Size size)
        {
            return ((_width >= size._width) && (_height >= size._height));
        }

        public Size Multiply(double amount)
        {
            return new Size(_width*amount, _height*amount);
        }

        public Size Multiply(double width, double height)
        {
            return new Size(_width*width, _height*height);
        }

        public Size Divide(double amount)
        {
            return new Size(_width/amount, _height/amount);
        }

        public static Size operator *(Size left_size, double right_size)
        {
            return left_size.Multiply(right_size);
        }

        public static Size operator /(Size left_size, double right_size)
        {
            return left_size.Divide(right_size);
        }

        public Size Add(Point point)
        {
            return new Size(_width + point.X, _height + point.Y);
        }

        public Size Add(Size size)
        {
            return new Size(_width + size._width, _height + size._height);
        }

        public Size Add(double width, double height)
        {
            return new Size(_width + width, _height + height);
        }

        public static Size operator +(Size size, Point point)
        {
            return size.Add(point);
        }

        public static Size operator +(Size left_size, Size right_size)
        {
            return left_size.Add(right_size);
        }

        /// <summary>
        /// Given the aspect ratio of inputsize, will return
        /// a size that fits within maxsize AND that maintains the
        /// original aspect ration
        /// </summary>
        /// <param name="inputsize"></param>
        /// <param name="maxsize"></param>
        /// <returns></returns>
        public static Size FitToArea(Size inputsize, Size maxsize)
        {
            double f = GetFitToAreaScalingFactor(inputsize._width, inputsize._height, maxsize._width, maxsize._height);
            var output = new Size(inputsize._width*f, inputsize._height*f);
            return output;
        }

        public static double GetFitToAreaScalingFactor(Size size, Size max)
        {
            return GetFitToAreaScalingFactor(size._width, size._height, max._width, max._height);
        }

        public static double GetFitToAreaScalingFactor(double inputwidth, double inputheight, double maxwidth,
                                                       double maxheight)
        {
            if ((inputwidth <= 0) || (inputheight <= 0) || (maxwidth <= 0) || (inputheight <= 0))
            {
                throw new System.ArgumentException("one of the input heightsor widths is negative");
            }

            double input_aspect_ratio = inputheight/inputwidth;
            double bounding_apect_ratio = maxheight/maxwidth;
            double scaling_factor;

            if (input_aspect_ratio <= bounding_apect_ratio)
            {
                scaling_factor = maxwidth/inputwidth;
            }
            else
            {
                scaling_factor = maxheight/inputheight;
            }

            // check to make sure scaling factor can be used
            if (inputwidth*scaling_factor > maxwidth + 1)
            {
                throw new DrawingException("incorrectly calculated scaling factor for width");
            }
            if (inputheight*scaling_factor > maxheight + 1)
            {
                throw new DrawingException("incorrectly calculated scaling factor for height");
            }

            return scaling_factor;
        }

        public static Size RoundToNearest(Size size, Size amount)
        {
            var new_size = new Size(
                DrawingUtil.Round(size._width, System.MidpointRounding.ToEven, amount._width),
                DrawingUtil.Round(size._height, System.MidpointRounding.ToEven, amount._height));
            return new_size;
        }

        public static Size RoundUpToNearest(Size size, Size amount)
        {
            var new_size = new Size(
                DrawingUtil.RoundUp(size._width, amount._width),
                DrawingUtil.Round(size._height, System.MidpointRounding.ToEven, amount._height));
            return new_size;
        }

        public static Size SnapToNearestValue(Size size, double width, double height)
        {
            return new Size(
                DrawingUtil.Round(size._width, width),
                DrawingUtil.Round(size._height, height));
        }

        public static Size SnapToNearestValue(Size size, Size snapsize)
        {
            return new Size(DrawingUtil.Round(size._width, snapsize._width),
                           DrawingUtil.Round(size._height, snapsize._height));
        }
    }
}