using System;
using System.Collections.Generic;

namespace Viziblr.Drawing
{
    public static partial class DrawingUtil
    {
        public static Point GetPointAtRadius(Point origin, double angle, double radius)
        {
            var new_point = new Point(radius * Math.Cos(angle),
                                      radius * Math.Sin(angle));
            new_point = origin + new_point;
            return new_point;
        }


        public static Point Max(Point a, Point b)
        {
            return new Point(Math.Max(a.X, b.X),
                             Math.Max(a.Y, b.Y));
        }

        public static Point Min(Point a, Point b)
        {
            return new Point(Math.Min(a.X, b.X),
                             Math.Min(a.Y, b.Y));
        }

        public static Size Max(Size a, Size b)
        {
            return new Size(Math.Max(a.Width, b.Width),
                            Math.Max(a.Height, b.Height));
        }

        public static Size Min(Size a, Size b)
        {
            return new Size(Math.Min(a.Width, b.Width),
                            Math.Min(a.Height, b.Height));
        }


        public static bool IsHorizontal(DirectionRelative direction)
        {
            return (direction == DirectionRelative.Left || direction == DirectionRelative.Right);
        }



        public static Point RotateAroundOrigin(Point p1, double theta)
        {
            double nx = (Math.Cos(theta)*p1.X) - (Math.Sin(theta)*p1.Y);
            double ny = (Math.Sin(theta)*p1.X) + (Math.Cos(theta)*p1.Y);
            return new Point(nx, ny);
        }

        internal static IEnumerable<ArcSegment> subdivide_arc_nicely(double start_angle, double end_angle)
        {
            // TODO: Should calculate number of subarcs without resorting to an enumeration
            // TODO: add test cases

            if (start_angle > end_angle)
            {
                throw new ArgumentException("end_angle must be less than start angle", "end_angle");
            }

            // the original purpose of this method is to break apart arcs > 90 degrees into smaller sub-arcs of 90 or less
            // the current implementation does that but also fractures the arc on 0,90,180,270, etc. degrees even if the arc length is <90
            // example (85,110) -> (85,90) & (90,110)

            double right_angle = Math.PI/2;

            var cur = new ArcSegment(start_angle, end_angle);
            while (true)
            {
                double temp = Math.Floor(cur.begin/right_angle);
                double cut_angle = (temp + 1)*right_angle;

                if ((cur.begin < cut_angle) && (cut_angle < cur.end))
                {
                    yield return (new ArcSegment(cur.begin, cut_angle));
                    cur = new ArcSegment(cut_angle, cur.end);
                }
                else
                {
                    yield return cur;
                    break;
                }
            }
        }

        /// <summary>
        /// Helper function to calculate distance between two points
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static double Distance(double a, double b)
        {
            return Math.Sqrt((a*a) + (b*b));
        }

        /// <summary>
        /// Helper function to calculate distance between two points
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <returns></returns>
        public static double Distance(double a, double b, double c)
        {
            return Math.Sqrt((a*a) + (b*b) + (c*c));
        }

        public static double RadiansToRange_0_1(double radians)
        {
            return WrapAngle_0_1(radians/(Math.PI*2));
        }

        /// <summary>
        /// Combines two input numbers in some proportion. 
        /// ratio=0.0 the first number is not used at all, 
        /// ratio=0.5 they are weight equally
        /// ratio=1.0 the first number completely dominates the value
        /// </summary>
        /// <param name="val0"></param>
        /// <param name="val1"></param>
        /// <param name="ratio"></param>
        /// <returns></returns>
        public static double Blend_0_1(double val0, double val1, double ratio)
        {
            var cratio = ClampToRange_0_1(ratio);
            var v0 = val0 * cratio;
            var v1 = val1 * (1.0 - cratio);
            return v0 + v1;
        }

        /// <summary>
        /// Checks if a value is in a range (inclusive)
        /// </summary>
        /// <param name="val"></param>
        /// <param name="min"></param>
        /// <param name="max"></param>
        /// <returns></returns>
        public static bool IsInRange(double val, double min, double max)
        {
            return ((min <= val) && (val <= max));
        }

        /// <summary>
        /// Checks if a value is in the range 0.0 to 1.0 inclusive
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        public static bool IsInRange_0_1(double val)
        {
            return IsInRange(val, 0.0, 1.0);
        }

        /// <summary>
        /// Given an input value will force the value to fit within the range (min,max) inclusive
        /// </summary>
        /// <param name="v"></param>
        /// <param name="min"></param>
        /// <param name="max"></param>
        /// <returns></returns>
        public static double ClampToRange(double v, double min, double max)
        {
            if (v < min)
            {
                v = min;
            }
            else if (v > max)
            {
                v = max;
            }
            return v;
        }

        /// <summary>
        /// Given an input value, will limit it to the range 0.0 and 1.0 inclusive
        /// </summary>
        /// <param name="v"></param>
        /// <returns>the clamped value</returns>
        public static double ClampToRange_0_1(double v)
        {
            return ClampToRange(v, 0.0, 1.0);
        }

        /// <summary>
        /// This is a variant of mod that wraps the mod result to avoid negative results. this is what Python's mod operator does
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        private static double mod_wrap_angle(double x, double y)
        {
            if (y == 0)
            {
                throw new DivideByZeroException();
            }

            double r = x % y;
            if (r > 0 && y < 0)
            {
                r = r + y;
            }
            else if (r < 0 && y > 0)
            {
                r = r + y;
            }
            return r;
        }

        /// <summary>
        /// wraps a number around so that it always fits between 0.0 and 1.0. negative numbers will wrap around to the correct positive number
        /// </summary>
        /// <remarks>
        /// if the input number is already in the range, no change will occur
        /// </remarks>
        /// <param name="v">input value </param>
        /// <returns>the wrapped number</returns>
        public static double WrapAngle_0_1(double v)
        {
            const double min = 0.0;
            const double max = 1.0;
            if (IsInRange(v, min, max))
            {
                // the number is already in the range so do nothing
                return v;
            }
            return mod_wrap_angle(v, max);
        }

        /// <summary>
        /// Returns the maximum value of three numbers
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <returns></returns>
        public static double Max(double a, double b, double c)
        {
            return (Math.Max(Math.Max(a, b), c));
        }

        /// <summary>
        /// Returns the minimum value of three numbers
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <param name="c"></param>
        /// <returns></returns>
        public static double Min(double a, double b, double c)
        {
            return (Math.Min(Math.Min(a, b), c));
        }

        public static double Round(double val, double snap_val)
        {
            return Round(val, MidpointRounding.AwayFromZero, snap_val);
        }

        /// <summary>
        /// rounds val to the nearest fractional value 
        /// </summary>
        /// <param name="val">the value tp round</param>
        /// <param name="rounding">what kind of rounding</param>
        /// <param name="frac"> round to this value (must be greater than 0.0)</param>
        /// <returns>the rounded value</returns>
        public static double Round(double val, MidpointRounding rounding, double frac)
        {
            /*
            if (frac <= 0)
            {
                throw new ArgumentOutOfRangeException("frac","must be greater than or equal to 0.0");
            }*/
            double retval = Math.Round((val / frac), rounding) * frac;
            return retval;
        }

        public static double RoundUp(double v, double amount)
        {
            const MidpointRounding rounding = MidpointRounding.ToEven;
            var result = Round(v + (amount / 2.0), rounding, amount);
            return result;
        }

        public static Rectangle? TryGetBoundingBox(IEnumerable<Point> points)
        {
            int count = 0;
            Point min = default(Point);
            Point max = default(Point);
            int point_index = 0;
            foreach (var point in points)
            {
                if (point_index == 0)
                {
                    min = point;
                    max = point;
                }
                else
                {
                    min = Min(min, point);
                    max = Max(max, point);
                }
                count++;
                point_index++;
            }
            if (count > 0)
            {
                return new Rectangle(min.X, min.Y, max.X, max.Y);
            }
            return null;
        }

        public static Rectangle? TryGetBoundingBox(IEnumerable<Rectangle> rects)
        {
            int count = 0;
            Rectangle old_val = default(Rectangle);

            int rect_index = 0;
            foreach (var rect in rects)
            {
                if (rect_index == 0)
                {
                    old_val = rect;
                }
                else
                {
                    old_val = Rectangle.FromBoundingBox(old_val, rect);
                }
                count++;
                rect_index++;
            }
            if (count > 0)
            {
                return old_val;
            }
            return null;
        }
    }
}